#! /usr/bin/python3.9

"""
Mahyar@Mahyar24.com, Thu 19 Aug 2021.
"""


from Cython.Build import cythonize
from setuptools import Extension
from setuptools.command.build_ext import build_ext

ext_modules = cythonize(
    [
        Extension(
            name="findsub.core.algo",
            include_dirs=["findsub/core"],
            sources=["findsub/core/algo.pyx"],
        ),
    ]
)


class BuildFailed(Exception):
    pass


class ExtBuilder(build_ext):
    def run(self):
        try:
            super().run()
        except (DistutilsPlatformError, FileNotFoundError):
            raise BuildFailed("File not found. Could not compile Cython extensions.")

    def build_extension(self, ext):
        try:
            super().build_extension(ext)
        except (CCompilerError, DistutilsExecError, DistutilsPlatformError, ValueError):
            raise BuildFailed("Could not compile Cython extensions.")


def build(setup_kwargs):
    """
    This function is mandatory in order to build the extensions.
    """
    setup_kwargs.update(
        {"ext_modules": ext_modules, "cmdclass": {"build_ext": ExtBuilder}}
    )



#! /usr/bin/python3.9

"""
Mahyar@Mahyar24.com, Thu 19 Aug 2021.
"""

from datetime import timedelta

def match(
    base: list[tuple[timedelta, timedelta]], other: list[tuple[timedelta, timedelta]]
) -> float: ...



#! /usr/bin/python3.9

"""
This module's goal is to get a movie filename, clean it, and by using IMDB api return the
official title of movie.
Compatible with python3.9+.
`IMDbPY` library is required. -> https://pypi.org/project/IMDbPY/
In case of failure, it will raise a ValueError.
Compatible with python3.9+.
Mahyar@Mahyar24.com, Thu 19 Aug 2021.
"""

import pathlib
import string
from datetime import date
from hashlib import md5

from imdb import IMDb  # type: ignore


class Movie:
    """
    Creating an instance of movie and process the data of it base on IMDB API.
    """

    def __init__(self, file: pathlib.Path) -> None:
        self.path = file.absolute()
        self.filename = self.path.name
        self.filename_only = self.path.stem
        self.mime = self.path.suffix
        self.dir = self.path.parent.absolute()
        self.filename_hash = md5(self.filename.encode("utf-8")).hexdigest()
        self.suggested_separator = "."
        self.suggested_year = ""
        self.imdb = IMDb()

    def _find_separator(self) -> str:
        """
        Checking for most used punctuation or whitespace to suggest it as separator.
        """
        separators = frozenset(string.punctuation + string.whitespace)
        occurrences = []
        for character in separators:
            occurrences.append((character, self.filename_only.count(character)))
        return max(occurrences, key=lambda x: x[1])[0]

    @staticmethod
    def _find_year(split_filename: list[str]) -> str:
        """
        Checking for existence of year in filename. if none, '' will return.
        """
        for part in split_filename:
            if part.isdigit():
                if date.today().year >= int(part) >= 1890:  # Sensible year for a movie.
                    return part
        return ""

    def clean_filename(self) -> str:
        """
        We have an assumption that file names are in this kind of patterns:
            'Borat.Subsequent.Moviefilm.2020.1080p.WEB-DL.RARBG.DigiMoviez.mkv'
            'Hellboy II The Golden Army 2008 720p (MarzFun.ir).mp4'
        NO series support yet!
        """
        self.suggested_separator = self._find_separator()
        split_filename = self.filename_only.split(self.suggested_separator)
        self.suggested_year = self._find_year(split_filename)
        if self.suggested_year:
            result = (
                " ".join(split_filename[: split_filename.index(self.suggested_year)])
                + f" ({self.suggested_year})"
            )
        else:
            name = []
            for part in split_filename:
                if part.replace(r"'", "").isalpha():
                    name.append(part)
                else:
                    break
            result = " ".join(name)

        if result:
            return result
        raise ValueError(f"Cannot clean the {self.filename_only}!")

    def search(self) -> tuple[str, str]:
        """
        Based on a clean name, search in IMDB api for title. raise ValueError if nothing pop out.
        """
        clean_filename = self.clean_filename()
        result = self.imdb.search_movie(clean_filename)
        if result:
            suggested_movie = result[0]
            return suggested_movie.data["title"], suggested_movie.data["year"]
        raise ValueError(f"Cannot find: {clean_filename!r}")

    def __repr__(self) -> str:
        return f"Movie(Path={self.path!r}, Name={self.filename_only!r})"



#! /usr/bin/python3.9

"""
This module's goal is to calculate mutual percentage of time between two data structure
based on `subtitle.py` module. It will show how much they are potentially synchronous.
`tqdm` library is required. -> https://pypi.org/project/tqdm/
Compatible with python3.9+.
Mahyar@Mahyar24.com, Thu 19 Aug 2021.
"""

from concurrent.futures import ProcessPoolExecutor, as_completed
from datetime import timedelta

from tqdm import tqdm  # type: ignore

from .core import match


def match_all(
    movie_time: list[tuple[int, int]],
    sub_times: dict[str, list[tuple[timedelta, timedelta]]],
) -> dict[str, float]:
    """
    See match function docstring. matching concurrently and sorting the result.
    """
    result = {}
    with ProcessPoolExecutor() as executor:
        tasks = {executor.submit(match, movie_time, v): k for k, v in sub_times.items()}
        for task in tqdm(
            as_completed(tasks.keys()),
            desc="Matching Subtitles",
            total=len(sub_times),
            bar_format="{desc}: {bar} {n_fmt}/{total_fmt} {percentage:3.0f}%",
        ):
            result[tasks[task]] = task.result()
    return dict(sorted(result.items(), key=lambda item: item[1], reverse=True))



#! /usr/bin/python3.9

"""
This module's goal is to make a data structure of when there is a human speech in the movie.
`webrtcvad` library is required. -> https://pypi.org/project/webrtcvad/
`FFmpeg` is required. -> https://www.ffmpeg.org/
`FFprobe` is required. -> https://ffmpeg.org/ffprobe.html
Some functions here are copied from https://github.com/wiseman/py-webrtcvad. (MIT License)
Compatible with python3.9+.
Mahyar@Mahyar24.com, Thu 19 Aug 2021.
"""

import contextlib
import wave
from pathlib import Path

import webrtcvad  # type: ignore

from .ffmpeg import RATES


def generate_chunk(file: Path, frame_duration_ms: int, sample_rate: int) -> list[bool]:
    """
    Slicing was to chunk of data based on frame duration.
    """
    vad = webrtcvad.Vad()
    vad.set_mode(0)

    base_ms = []
    num_frames = int(sample_rate * (frame_duration_ms / 1000.0))
    with contextlib.closing(wave.open(str(file), "rb")) as wav_file:
        with contextlib.suppress(Exception):
            while chunk := wav_file.readframes(num_frames):
                base_ms.append(vad.is_speech(chunk, sample_rate))

    return base_ms


def assert_wave(file: Path) -> int:
    """
    asserting qualities, returning sample rate.
    """
    with contextlib.closing(wave.open(str(file), "rb")) as wav_file:
        num_channels = wav_file.getnchannels()

        assert (
            num_channels == 1
        ), f"Number of Channels must be one, not {num_channels!r}."

        sample_width = wav_file.getsampwidth()
        assert sample_width == 2, f"Width must be two, not {sample_width!r}."

        sample_rate = wav_file.getframerate()
        assert (
            sample_rate in RATES
        ), f"Sample Rate must be in {RATES}, not {sample_rate!r}."

        return sample_rate


def make_base(
    file: Path, millisecond: int = 20, threshold: float = 0.85
) -> list[tuple[int, int]]:
    """
    We will use only this function externally.
    Make a timeline structure of when there is speech. For increasing the speed,
    this function based on threshold will decide that in every one second is
    there a human speech or not.
    """
    rate = assert_wave(file)

    unit = 1_000 // millisecond

    base_ms = generate_chunk(file, millisecond, rate)
    base_s = [base_ms[i : i + unit] for i in range(0, len(base_ms), unit)]
    base = []
    for i, second in enumerate(base_s):
        if sum(second) / len(second) > threshold:
            base.append(
                (
                    i,
                    i + 1,
                )
            )
    return base



#! /usr/bin/python3.9

"""
This module's goal is to extract audio of the movie in a way that satisfy the
technical requirements of `webrtcvad`.
`FFmpeg` is required. -> https://www.ffmpeg.org/
`FFprobe` is required. -> https://ffmpeg.org/ffprobe.html
Mahyar@Mahyar24.com, Thu 19 Aug 2021.
"""

import bisect
import shutil
import subprocess
from pathlib import Path
from typing import Literal, Union

from .movie import Movie

RATES = (8_000, 16_000, 32_000, 48_000)


class FFmpegError(BaseException):
    """We Should terminate the program if this error is raised."""


def find_sample_rate(rate: int) -> int:
    """
    Find best sample rate. Closest lower one.
    """
    index = bisect.bisect(RATES, rate)
    if index == 0:
        return RATES[0]
    return RATES[index - 1]


def suggest_sample_rate(movie: Movie) -> Union[int, Literal[False]]:
    """
    If sample_rate was in (8000, 16000, 32000, 48000)Hz,
    we return a False and doesn't resample the rate; otherwise
    we will choose closest lower sample rate.
    """
    assert shutil.which("ffprobe") is not None, "Cannot find FFprobe."

    command = (
        f"ffprobe -hide_banner -select_streams a:0 -show_entries"
        f" stream=sample_rate -of default=noprint_wrappers=1:nokey=1 '{movie.path}'"
    )

    try:
        sample_rate = subprocess.check_output(
            command, shell=True, stderr=subprocess.DEVNULL, text=True
        ).strip()
    except subprocess.CalledProcessError:
        print("FFprobe cannot extract audio sample rate! It will set to 16,000.")
        return 16_000
    else:
        rate = int(float(sample_rate))
        if rate in RATES:
            return False
        return find_sample_rate(rate)


def extract_audio(movie: Movie, cached_audio: Path) -> None:
    """
    Extracting audio of the movie with help of `FFmpeg`.
    16-bit. Mono. 32,000 Hz. Wav.
    """

    destination = movie.dir / f".{movie.filename_hash}_audio.wav"

    assert shutil.which("ffmpeg") is not None, "Cannot find FFmpeg."
    if rate := suggest_sample_rate(movie):
        command = (
            f"ffmpeg -y -i '{movie.path}' -map a:0 -ar {rate} -acodec pcm_s16le"
            f" -ac 1 '{destination}'"
        )
    else:
        command = f"ffmpeg -y -i '{movie.path}' -map a:0 -acodec pcm_s16le -ac 1 '{destination}'"

    return_code = subprocess.call(
        command, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
    )

    if return_code:  # Error
        msg = "FFmpeg cannot extract audio! "
        if ":" in str(movie.path):
            msg += 'maybe because there is a ":" in filename!'

        raise FFmpegError(msg)

    # In case of SIGKILL the file doesn't exist!
    if destination.is_file():
        destination.rename(cached_audio)



#! /usr/bin/python3.9

"""
This module's goal is to unzip, removes duplicate and convert subtitles to
UTF-8 and remove unnecessary junks.
Compatible with python3.9+. No third-party library is required, but having a dependency for
`Convert.sh` and `iconv` command underneath.
`iconv` is required. -> https://www.gnu.org/software/libiconv/
also Bash is in need!
Compatible with python3.9+.
Mahyar@Mahyar24.com, Thu 19 Aug 2021.
"""

import hashlib
import shutil
import subprocess
import zipfile
from concurrent.futures import ALL_COMPLETED, ProcessPoolExecutor, wait
from pathlib import Path
from typing import Optional


def hash_subtitles(directory: Path) -> None:
    """
    Hash the subtitles content by using md5 and then renaming it to the hexdigest of hash.
    """
    for item in directory.iterdir():
        if item.is_file() and item.name.endswith(".srt"):
            with open(item, "rb") as subtitle:
                data = subtitle.read()
            hash_name = hashlib.md5(data).hexdigest() + ".srt"
            item.rename(directory / hash_name)


def delete_bad_files(directory: Path) -> None:
    """
    Delete any non-srt files.
    """
    for item in directory.iterdir():
        if item.is_file() and not item.name.endswith(".srt"):
            item.unlink(missing_ok=True)


def unzip_remove(zip_file: Path) -> Optional[Path]:
    """
    Unzip, extract all and removing zip file. If extraction fails, remove
    the zip file and return None, else, delete non-srt files and change
    the name of subtitles to their md5 hash hexdigest.
    """
    directory = zip_file.parent / zip_file.stem
    try:
        with zipfile.ZipFile(zip_file, "r") as file:
            # Must extract every zip file to one directory otherwise maybe some
            # subtitle with duplicate name and different content gets remove.
            file.extractall(directory)
    except zipfile.BadZipfile:
        zip_file.unlink()
        return None
    else:

        zip_file.unlink()
        delete_bad_files(directory)
        hash_subtitles(directory)

        return directory


def unzip_all_and_hash(zip_files: list[Path]) -> list[Path]:
    """
    Unzip and calculate hash for renaming concurrently.
    Returning list of successfully processed subtitles.
    """
    with ProcessPoolExecutor() as executor:
        tasks = [
            executor.submit(unzip_remove, zip_file)
            for zip_file in zip_files
            if zip_file.name.endswith(".zip")
        ]
        wait(tasks, return_when=ALL_COMPLETED)
    return [result for task in tasks if (result := task.result()) is not None]


def iconv_subtitles(directory: Path) -> None:
    """
    Converting non UTF-8 srt files to UTF-8. Based on `Convert.sh`.
    Because we are in subtitle directory, we must run shell script
    with leading two dots. (../Sample.sh)
    """
    script_path = Path(__file__).parent / "scripts/Convert.sh"
    subprocess.call(
        f'{script_path} "{directory}"',
        shell=True,
        stdout=subprocess.DEVNULL,
    )


def move_up(directories: list[Path]) -> None:
    """
    Moving up subtitles and delete the empty directory.
    """
    for directory in directories:
        for item in directory.iterdir():
            if item.is_file():
                item.rename(item.absolute().parents[1] / item.name)
            else:
                shutil.rmtree(item)
        directory.rmdir()


def prepare_files(directory: Path) -> None:
    """
    We will use only this function externally.
    Check other functions docstring.
    """
    files = list(directory.iterdir())
    dirs = unzip_all_and_hash(files)
    move_up(dirs)
    iconv_subtitles(directory)



#! /usr/bin/python3.9

"""
Some tiny functions to use in `__main__.py`.
Compatible with python3.9+.
Mahyar@Mahyar24.com, Thu 19 Aug 2021.
"""

import json
import math
import os
import shutil
from pathlib import Path

from .movie import Movie


def emergency_cleanup(movie: Movie) -> None:
    """
    Clean all cached and unused file and directories in case of a sudden failure.
    """
    completed_audio = movie.dir / f".{movie.filename_hash}_audio_completed.wav"
    uncompleted_audio = movie.dir / f".{movie.filename_hash}_audio.wav"
    hidden_sub_dir = movie.dir / f".{movie.filename_hash}"

    completed_audio.unlink(missing_ok=True)
    uncompleted_audio.unlink(missing_ok=True)

    try:
        shutil.rmtree(hidden_sub_dir)
    except FileNotFoundError:
        pass


def clear(directory: Path, cached_audio: Path, remove: bool) -> None:
    """
    Remove the directory and audio file.
    """
    if remove:
        shutil.rmtree(
            directory
        )  # The directory should be empty but some cautious is not bad!

    cached_audio.unlink(missing_ok=True)


def find_zero_pad_number(length_of_subtitles: int) -> int:
    """
    Finding needed length of zeros. 123 files -> 1: 001; so it needs two leading zero.
    """
    return int(math.log10(length_of_subtitles)) + 1


def make_subs_dir(
    directory: Path, results: dict[str, float], move: bool = True
) -> None:
    """
    Make the Subs directory and rename subtitles based on coverage.
    """
    base_dir = directory.parent.absolute()
    subs = base_dir / "Subs"

    try:
        os.mkdir(subs)
    except FileExistsError:
        pass

    zero_pad_num = find_zero_pad_number(len(results))

    info: dict[str, dict[str, str]] = {
        "Subs": {},
        "FindSub": {
            "GitHub": "https://github.com/mahyar24/findsub",
            "PyPI": "https://pypi.org/project/findsub/",
            "E-Mail": "Mahyar@Mahyar24.com",
        },
    }

    for i, sub in enumerate(results.keys()):

        old_file = directory / sub
        new_name = f"{i + 1}".zfill(zero_pad_num) + ".srt"
        new_file = subs / new_name

        if move:
            old_file.rename(new_file)
        else:
            try:
                shutil.copy(old_file, new_file)
            except shutil.SameFileError:
                pass

        if results[sub] >= 0.0:  # If synchronous ratio became negative!
            print(f"{new_name}: {results[sub]:.2%}")
            info["Subs"][new_name] = f"{results[sub]:.2%}"

    with open(subs / "FindSub.json", "w", encoding="utf-8") as info_file:
        json.dump(info, info_file, indent=4)



#! /usr/bin/python3.9

"""
This module's goal is to download all subtitles of a suggested movie title
in a specific language from Subscene site.
Compatible with python3.9+.
`aiofiles` library is required. -> https://pypi.org/project/aiofiles/
`aiohttp` library is required. -> https://pypi.org/project/aiohttp/
`beautifulsoup4` library is required. -> https://pypi.org/project/beautifulsoup4/
`lxml` library is required. -> https://pypi.org/project/lxml/
`tqdm` library is required. -> https://pypi.org/project/tqdm/
Mahyar@Mahyar24.com, Thu 19 Aug 2021.
"""

import asyncio
import os
import shutil
from pathlib import Path
from typing import Optional

import aiofiles
import aiohttp
from bs4 import BeautifulSoup  # type: ignore
from tqdm import tqdm  # type: ignore

from .movie import Movie


class Downloader:
    """
    Download Subtitles from subscene.
    """

    SUBSCENE_URL = "https://subscene.com"

    def __init__(
        self,
        movie: Movie,
        lang: str,
        link: Optional[str] = None,
    ) -> None:
        self.lang = lang
        self.link = link
        self.movie = movie

    def suggest_link(self) -> None:
        """
        based on `filename.py` module. (see module's docstring)
        Return title and suggested Subscene link; If it fails for getting title,
        it will raise a ValueError.
        """
        try:
            name, year = self.movie.search()
        except ValueError as error:
            raise ValueError("IMDB API cannot find the name of this movie.") from error
        else:
            subscene_link = (
                self.SUBSCENE_URL
                + f"""/subtitles/{name.replace(":", "").replace(" ", "-")
                .replace("'", "").lower()}"""
            )
            self.link = subscene_link

            print(f"IMDB Search: {name!r} ({year}). Subscene link: {self.link!r}")

    async def get_content(self) -> str:
        """
        Get HTML of a link. It will raise a ValueError if respond wasn't ok.
        """
        async with aiohttp.ClientSession() as session:
            async with session.get(self.link) as resp:  # type: ignore
                if resp.ok:
                    return await resp.text()
                raise ValueError(
                    f"Cannot find: {self.link!r}: {resp.status!r}\n"
                    f"Please Specify the subscene link of this movie "
                    f"explicitly with help of -s/--subscene option."
                )

    async def get_subtitles_links(self) -> list[str]:
        """
        Scraping Subscene page and return a list os subtitle download pages
        with corresponding language.
        """
        content = await self.get_content()
        soup = BeautifulSoup(content, "lxml")
        links: list[str] = []
        for html_link in soup.find_all("a"):
            if rf"/{self.lang}/" in html_link["href"]:
                links.append(self.SUBSCENE_URL + html_link["href"])
        if not links:
            raise NotImplementedError(f"No subtitle with {self.lang!r} language found!")
        return links

    async def extract_dl_link(
        self, link: str, session: aiohttp.client.ClientSession
    ) -> Optional[str]:
        """
        Return download link from Subscene download page.
        """
        async with session.get(link) as resp:
            if resp.ok:
                content = await resp.text()
                soup = BeautifulSoup(content, "lxml")
                download_button = soup.find("a", {"id": "downloadButton"})
                return self.SUBSCENE_URL + download_button["href"]
        return None

    @staticmethod
    async def download_one(
        link: str,
        name: str,
        directory: Path,
        session: aiohttp.client.ClientSession,
    ) -> Optional[str]:
        """
        Download the subtitle, if response was not okay, return None.
        """
        async with session.get(link) as resp:
            if resp.ok:
                async with aiofiles.open(directory / name, "wb") as file:
                    while True:
                        chunk = await resp.content.read(1024)
                        if not chunk:
                            break
                        await file.write(chunk)
                return name
        return None

    async def download_all(self, links: list[str], directory: Path) -> list[str]:
        """
        Extract download links of subtitles and download them.
        """
        num = 0
        async with aiohttp.ClientSession() as session:
            extracting_download_links = [
                asyncio.create_task(self.extract_dl_link(link, session))
                for link in links
            ]
            downloading_files = []
            for extract_coro in tqdm(
                asyncio.as_completed(extracting_download_links),
                desc="Downloading Subtitles",
                total=len(links),
                bar_format="{desc}: {bar} {n_fmt}/{total_fmt} {percentage:3.0f}%",
            ):
                download_link = await extract_coro
                if download_link is not None:
                    num += 1
                    name = f"{num}.zip"
                    downloading_files.append(
                        asyncio.create_task(
                            self.download_one(download_link, name, directory, session)
                        )
                    )
            results = await asyncio.gather(*downloading_files)
        return [result for result in results if result is not None]

    async def _download(self) -> Path:
        """
        Main async entry point. It will make a directory with movie title (hidden) and cd to it.
        download the files and return list of their filenames and directory path.
        """
        directory = self.movie.dir / f".{self.movie.filename_hash}"

        if directory.is_dir():
            shutil.rmtree(directory)

        os.mkdir(directory)

        if self.link is None:
            self.suggest_link()

        links = await self.get_subtitles_links()
        await self.download_all(links, directory)
        return directory

    def download(self) -> Path:
        """
        Public Method for downloading the subtitles.
        """
        return asyncio.run(self._download())



#! /usr/bin/python3.9

"""
This module's goal is to make a data structure for subtitle times
Compatible with python3.9+.
`srt` library is required. -> https://pypi.org/project/srt/
Mahyar@Mahyar24.com, Thu 19 Aug 2021.
"""

import datetime
from concurrent.futures import ALL_COMPLETED, ProcessPoolExecutor, wait
from pathlib import Path

import srt  # type: ignore


def extract_subtitle_time(
    file_name: Path,
) -> list[tuple[datetime.timedelta, datetime.timedelta]]:
    """
    Making a data structure for times when there is a dialog.
    ---> [(start1, end1), (start2, end2), ...]
    In case of error, function will catch the exception and return empty list.
    """
    try:
        with open(file_name, "r", encoding="utf-8") as file:
            data = file.read().strip()
        sub = srt.parse(data)
        times = [(item.start, item.end) for item in sub]
    except (UnicodeDecodeError, srt.SRTParseError):
        return []
    else:
        return times


def extract_subtitle_times(
    directory: Path,
) -> dict[str, list[tuple[datetime.timedelta, datetime.timedelta]]]:
    """
    Making data structure for all subtitles concurrently.
    Check extract_subtitle_time function's docstring.
    """
    subtitles = list(directory.iterdir())
    tasks = {}
    with ProcessPoolExecutor() as executor:
        for subtitle in subtitles:
            if subtitle.name.endswith(".srt"):
                tasks[subtitle.name] = executor.submit(
                    extract_subtitle_time,
                    subtitle.absolute(),
                )
        wait(tasks.values(), return_when=ALL_COMPLETED)

    result = {}
    for key, value in tasks.items():
        if val := value.result():
            result[key] = val

    if result:
        return result
    raise UnicodeError("Cannot read any of the subtitles.")



#! /usr/bin/python3.9

"""
This module's goal is parsing the cli options and arguments.
Compatible with python3.9+.
Mahyar@Mahyar24.com, Thu 19 Aug 2021.
"""

import argparse
import json
import os
import pathlib
import pkgutil
import textwrap


def find_language(code: str) -> str:
    """
    Reading langs.json and trying to find desired language
    by checking two-letter codes base on ISO 639-1.
    Additionally, "bz" -> "brazillian-portuguese".
    """
    if (langs_file := pkgutil.get_data(__name__, "data/langs.json")) is None:
        raise FileNotFoundError("langs.json does not exist")

    langs = json.loads(langs_file.decode("utf-8"))

    for key, value in langs.items():
        if code.lower() == key:
            result = value["name"].split()[0]
            print(f'{code!r} -> {value["name"]!r}.')
            return result.lower()

    raise ValueError(f"{code!r} not found!")


def parsing_args() -> argparse.Namespace:
    """
    Parsing the passed arguments, read help (-h, --help) for further information.
    """
    parser = argparse.ArgumentParser(
        epilog=textwrap.dedent(
            """
            Written by: Mahyar Mahdavi <Mahyar@Mahyar24.com>.
            Source Code: <https://github.com/mahyar24/findsub>.
            PyPI: <https://pypi.org/project/findsub/>
            Reporting Bugs and PRs are welcomed. :)
            """
        )
    )
    group_link_dir = (
        parser.add_mutually_exclusive_group()
    )  # Link or directory, not both!

    parser.add_argument(
        "file", help="Select desired movie.", type=lambda x: pathlib.Path(x).absolute()
    )

    parser.add_argument(
        "-l",
        "--language",
        default=os.environ.get("FINDSUB_LANG", "en"),
        type=find_language,
        help="Two letter code for desired subtitle's language. (ISO 639-1) [Additionally: 'bz' -> "
        "'brazillian-portuguese']",
    )

    group_link_dir.add_argument(
        "-s", "--subscene", help="URL of subscene for the movie."
    )

    group_link_dir.add_argument(
        "-d",
        "--subtitles-directory",
        type=lambda x: pathlib.Path(x).absolute(),
        help="Check against already present subtitles in a directory.",
    )

    parser.add_argument(
        "-b",
        "--synced-subtitle",
        type=lambda x: pathlib.Path(x).absolute(),
        help="If you already have a synced subtitle, use it as base. (alot faster)",
    )

    parser.add_argument(
        "-a",
        "--audio",
        type=lambda x: pathlib.Path(x).absolute(),
        help="If extracted audio is available, use the path to it to speed up the program.",
    )

    return parser.parse_args()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['findsub', 'findsub.core']

package_data = \
{'': ['*'], 'findsub': ['data/*', 'scripts/*']}

install_requires = \
['Cython>=0.29.26,<0.30.0',
 'IMDbPY>=2021.4.18,<2022.0.0',
 'aiofiles>=0.8.0,<0.9.0',
 'aiohttp[speedups]>=3.8.1,<4.0.0',
 'beautifulsoup4>=4.10.0,<5.0.0',
 'lxml>=4.7.1,<5.0.0',
 'setuptools>=59.6.0,<60.0.0',
 'srt>=3.5.0,<4.0.0',
 'tqdm>=4.62.3,<5.0.0',
 'webrtcvad>=2.0.10,<3.0.0',
 'wheel>=0.37.0,<0.38.0']

entry_points = \
{'console_scripts': ['findsub = findsub.__main__:run']}

setup_kwargs = {
    'name': 'findsub',
    'version': '0.7.1',
    'description': 'Finding and ranking subtitles from subscene by how much they are synced.',
    'long_description': '# FindSub\n\nFindSub is an Application for automatically downloading and ranking subtitles based on how much they are synced\nfrom [subscene](https://subscene.com/).\n**Compatible With Unix Systems.**\n\n\n## Installation\n\nRecommended way is by using [pipx](https://github.com/pypa/pipx).\nFirst, install pipx:\n\n```bash\nsudo apt install pipx\n```\n\nThen install FindSub by using the command below:\n\n```bash\npipx install --python python3.9 findsub\n```\n\nP.S.: FindSub is supported by python3.9+. Also, GCC is required for compiling C extensions.\nFFmpeg and FFprobe are also needed for extracting of movie\'s audio.\n\n# Basic Usage and explanation. (Must read!)\n\n```bash\nfindsub The.French.Dispatch.2021.1080p.WEB-DL.x264.6CH-Pahe.FilmBan.mkv\n```\n→ First, findsub try to guess the name of the movie based on the filename and with the help of IMDB. \nSimultaneously it will extract the movie\'s audio by using "**FFmpeg**"  and "**FFprobe**". \nIf a cached extracted audio is present in the same directory, findsub will skip this stage and use it instead.\nAt the same time, findsub tries to download the subtitles from subscene in the desired language.\nYou can use -l/--language to select a language (ISO 639-1). if nothing is specified with this option, \n**findsub will try to use "FINDSUB_LANG" environment variable** and if nothing is set, it will use "En" as default.\nSometimes, especially when the original movie name is not in English, findsub cannot find the paired subscene page, \nand you should manually set the subscene link of the page with the help of -s/--subscene option.\n\n\n# Advance Usage\n## Languages \n\n```bash\nfindsub The.French.Dispatch.2021.1080p.WEB-DL.x264.6CH-Pahe.FilmBan.mkv --language "fa";\n```\nOR\n```\nexport FINDSUB_LANG="fa"; findsub The.French.Dispatch.2021.1080p.WEB-DL.x264.6CH-Pahe.FilmBan.mkv\n```\n→ With these two approaches, you can set the language.\nYou must use two-letter codes based on (ISO 639-1). \nP.S.: **exceptionally, use "bz" for Brazillian Portuguese.**\n\n## Speeding up\n```bash\nfindsub The.French.Dispatch.2021.1080p.WEB-DL.x264.6CH-Pahe.FilmBan.mkv --audio ./already_extraced_audio.wav\n```\nOR\n```bash\nfindsub The.French.Dispatch.2021.1080p.WEB-DL.x264.6CH-Pahe.FilmBan.mkv --synced-subtitle ./synced_sub.srt\n```\n→ use an already extracted audio or a sync subtitle to speed up the program.\n```bash\nfindsub The.French.Dispatch.2021.1080p.WEB-DL.x264.6CH-Pahe.FilmBan.mkv --subtitles-directory downloaded_sub/\n```\n→ Skip downloading subtitles and rank the subtitles within the mentioned directory.\n\n## -s/--subscene\n```bash\nsubfinder The_Sea_Inside_2004_720p_BrRip_YIFY.mkv -s https://subscene.com/subtitles/the-sea-inside-mar-adentro\n```\n→ Sometimes FindSub cannot find the subscene page for a movie, in that case, you should manually pass the link to it.\n- check `findsub --help` for more info.\n\n## Dependencies\n**FFmpeg, FFprobe, Bash, and Iconv** are needed in most cases.\n\n\n## Issues\nFindsub currently doesn\'t support downloading subtitles for series episodes. \nAlso, it may not work very well with windows, but it should be usable.\n\n\n## Contributing\n    Written by: Mahyar Mahdavi <Mahyar@Mahyar24.com>.\n    License: GNU GPLv3.\n    Source Code: <https://github.com/mahyar24/FindSub>.\n    PyPI: <https://pypi.org/project/FindSub/>.\n    Reporting Bugs and P.R.s are welcomed. :)\n\n## License\n[GPLv3](https://choosealicense.com/licenses/gpl-3.0)\n',
    'author': 'Mahyar Mahdavi',
    'author_email': 'Mahyar@Mahyar24.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mahyar24/FindSub/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True

"""
Compatible with python3.9+ and Cython0.29.25+.
Mahyar@Mahyar24.com, Thu 19 Aug 2021.
"""

from cython.view cimport array as cy_array
from cpython.datetime cimport timedelta


cpdef double match(list base, list other):
    """
    Based on the data structure, calculate that how much of the time that there is
    some speech going on in base, there is a subtitle in other.
    """
    cdef:
        unsigned int base_len = len(base)
        unsigned int other_len = len(other)
        double matched = 0.0
        double base_total = 0.0
        double [:, :] c_base = cy_array(shape=(base_len, 2), itemsize=sizeof(double), format="d")
        double [:, :] c_other = cy_array(shape=(other_len, 2), itemsize=sizeof(double), format="d")
        unsigned int dialog, speech
        unsigned int i, ii


    for i in range(base_len):
        c_base[i, 0] = <double> base[i][0]
        c_base[i, 1] = <double> base[i][1]
        base_total += c_base[i, 1] - c_base[i, 0]

    for ii in range(other_len):
        c_other[ii, 0] = <double> other[ii][0].total_seconds()
        c_other[ii, 1] = <double> other[ii][1].total_seconds()

    for speech in range(base_len):
        for dialog in range(other_len):
            if c_base[speech, 0] > c_other[dialog, 1] or c_other[dialog, 0] > c_base[speech, 1]:  # Huge SpeedUP.
                continue

            if (
                c_other[dialog, 0] <= c_base[speech, 0] and c_other[dialog, 1] >= c_base[speech, 1]
            ):  # base is included in other completely
                matched += c_base[speech, 1] - c_base[speech, 0]
            elif (
                c_other[dialog, 0] >= c_base[speech, 0] and c_base[speech, 1] >= c_other[dialog, 1]
            ):  # other is included in base completely
                matched += c_other[dialog, 1] - c_other[dialog, 0]
            elif c_other[dialog, 0] <= c_base[speech, 0] <= c_other[dialog, 1] <= c_base[speech, 1]:  # partially
                matched += c_other[dialog, 1] - c_base[speech, 0]
            elif c_base[speech, 0] <= c_other[dialog, 0] <= c_base[speech, 1] <= c_other[dialog, 1]:  # partially
                matched += c_base[speech, 1] - c_other[dialog, 0]

    return matched / base_total



#! /usr/bin/python3.9

"""
This package's goal is to download all subtitles with a specific language
of a movie from Subscene.com and then check every one of them to ranks them
by synchronous.
How? It will download subtitles and simultaneously extract audio of the file
by using `FFMPEG`, after that we check when there is a possible human speech
and make a timeline of it. Finally, we check how much of a time when there is
a possible speech each of the subtitles has a text.

Required PyPI Packages:
    `aiofiles` library is required. -> https://pypi.org/project/aiofiles/
    `aiohttp` library is required. -> https://pypi.org/project/aiohttp/
    `beautifulsoup4` library is required. -> https://pypi.org/project/beautifulsoup4/
    `lxml` library is required. -> https://pypi.org/project/lxml/
    `webrtcvad` library is required. -> https://pypi.org/project/webrtcvad/
    `IMDbPY` library is required. -> https://pypi.org/project/IMDbPY/
    `srt` library is required. -> https://pypi.org/project/srt/
    `Cython` is required. -> https://pypi.org/project/Cython/
    `tqdm` library is required. -> https://pypi.org/project/tqdm/
Required External Tools:
    `FFmpeg` is required. -> https://www.ffmpeg.org/
    `FFprobe` is required. -> https://ffmpeg.org/ffprobe.html
    `iconv` is required. -> https://www.gnu.org/software/libiconv/
    And also Bash!

Some functions here are copied from https://github.com/wiseman/py-webrtcvad.

Usage:
    findsub <file>. -> makes a `Subs` folder and put ranked subtitles in it
    findsub <file> -a/--audio extracted_audio.wav -> same as last one but
        using already extracted audio. (faster!)
    findsub -l/--language en/english <file> -> getting english subtitles.
        default is set by "FINDSUB_LANG" environment variable otherwise "English".
    findsub -s/--subscene <subscene-link> <file> -> no link suggestion. (faster!)
    findsub -d/--subtitles-directory <path-of-downloaded-subtitles> <file> ->
        using already download subtitles.
Compatible with python3.9+.
Mahyar@Mahyar24.com, Thu 19 Aug 2021.
"""

import multiprocessing
import os
import signal
from pathlib import Path
from typing import Optional

from .clean import iconv_subtitles, prepare_files
from .cli import parsing_args
from .download import Downloader
from .ffmpeg import extract_audio
from .movie import Movie
from .pycore import match_all
from .pyvideo import make_base
from .subtitles import extract_subtitle_time, extract_subtitle_times
from .tools import clear, emergency_cleanup, make_subs_dir


def main(
    movie: Movie,
    language: str,
    audio: Optional[Path] = None,
    subscene: Optional[str] = None,
    subtitles_directory: Optional[Path] = None,
    synced_subtitle: Optional[Path] = None,
) -> None:
    """
    Main entry point. It should not be used within python code. Designed for CLI.
    """

    cached_audio = movie.dir / f".{movie.filename_hash}_audio_completed.wav"

    if synced_subtitle is None:
        if audio is None:  # Check for extracted audio file.
            if cached_audio.is_file():
                audio = cached_audio

        if audio is None:
            process = multiprocessing.Process(
                target=extract_audio, args=(movie, cached_audio), daemon=True
            )
            process.start()
            print("Audio extraction begins.")

    # If user already has a directory of subtitles, we must not move them to the Subs.
    move = True
    if subtitles_directory is None:
        subtitles_directory = Downloader(
            movie=movie, lang=language, link=subscene
        ).download()
        prepare_files(subtitles_directory)
    else:
        move = False
        iconv_subtitles(subtitles_directory)

    try:
        print("Examining subtitles.", end=" ", flush=True)
        sub_time_structures = extract_subtitle_times(subtitles_directory)
    except UnicodeError:
        clear(subtitles_directory, cached_audio, remove=move)
        raise
    else:
        print("Done.")

    if synced_subtitle is None:
        if audio is None:
            print("Waiting for audio extraction to finish.", end=" ", flush=True)
            # noinspection PyUnboundLocalVariable
            process.join()
            audio = cached_audio
            print("Done.")

        print("Voice Activity Detector started the analysis.", end=" ", flush=True)
        movie_time_structure = make_base(audio)
        print("Done.")
    else:
        temp_movie_time_structure = extract_subtitle_time(synced_subtitle)
        if temp_movie_time_structure:  # if it's not empty.
            movie_time_structure = [
                (
                    int(i[0].total_seconds()),
                    int(i[1].total_seconds()),
                )
                for i in temp_movie_time_structure
            ]
        else:
            clear(subtitles_directory, cached_audio, remove=move)
            raise UnicodeError(f"Cannot read '{synced_subtitle}'.")

    results = match_all(movie_time_structure, sub_time_structures)

    make_subs_dir(subtitles_directory, results, move=move)
    clear(subtitles_directory, cached_audio, remove=move)

    print("Done.")


def run():
    """
    EntryPoint of Application.
    """
    args = parsing_args()

    assert args.file.is_file(), f"Cannot find {args.file!r}"
    if args.audio is not None:
        assert args.file.is_file(), f"Cannot find {args.file!r}"
    if args.subtitles_directory is not None:
        assert (
            args.subtitles_directory.is_dir()
        ), f"Cannot find {args.subtitles_directory!r}"

    movie = Movie(args.file)

    try:
        main(
            movie=movie,
            language=args.language,
            audio=args.audio,
            subscene=args.subscene,
            subtitles_directory=args.subtitles_directory,
            synced_subtitle=args.synced_subtitle,
        )
    except BaseException as error:
        print(error)
        emergency_cleanup(movie)
        os.kill(os.getpid(), signal.SIGKILL)


if __name__ == "__main__":
    run()
